@@ -48,12 +48,15 @@ extern "C" {
 	void determinant( double A[], double *det_A, int *p );
 	
 	void select_edge( double rates[], int *index_selected_edge, double *sum_rates, int *qp );
-	
+
 	void select_multi_edges( double rates[], int index_selected_edges[], int *size_index, double *sum_rates, int *multi_update, int *qp );
 
 	void rates_bdmcmc_parallel( double rates[], int G[], int index_row[], int index_col[], int *sub_qp, double Ds[], double Dsijj[],
 					            double sigma[], double K[], int *b, int *p );
 
+	void rates_cbdmcmc_parallel( double rates[], int G[], int index_row[], int index_col[], int *sub_qp, double r_Ds[], double i_Ds[],
+				            double r_sigma[], double i_sigma[], double r_K[], double i_K[], int *b, int *p );
+
 	void log_H_ij( double K[], double sigma[], double *log_Hij, int *selected_edge_i, int *selected_edge_j,
 				   double Kj12[], double Kj12xK22_inv[], double K12[], double K12xK22_inv[], double K121[], 
 				   double sigmaj12[], double sigmaj22[], double sigma12[], double sigma22[], double sigma11_inv[], double sigma21xsigma11_inv[],
